import unittest
import os
import sys
import tempfile
import io
import contextlib
import shutil

# Add project root to sys.path to allow importing concat_all
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from concat_all import concat_all

class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()
        self.test_dir_path = self.test_dir.name

    def tearDown(self):
        self.test_dir.cleanup()

    def _create_files(self, file_dict):
        """
        Creates files and directories within the temporary test directory.
        file_dict: A dictionary where keys are relative file paths
                   and values are their content. Directories are created
                   implicitly. If content is None, an empty directory is created.
        """
        for rel_path, content in file_dict.items():
            full_path = os.path.join(self.test_dir_path, rel_path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            if content is not None:
                with open(full_path, 'w', encoding='utf-8') as f:
                    f.write(content)
            else: # content is None, ensure directory exists
                if not os.path.exists(full_path): # only if it's not implicitly created by another file
                    os.makedirs(full_path, exist_ok=True)


    def _run_concat_files(self, extensions="txt", output_file_name="concat_out.txt", **kwargs):
        """
        Runs concat_files and captures its stdout, then returns output file content and stdout.
        """
        output_file_path = os.path.join(self.test_dir_path, output_file_name)
        
        # Default kwargs for concat_files, can be overridden
        run_kwargs = {
            'dir_path': self.test_dir_path,
            'file_extensions': extensions,
            'output_file': output_file_path,
            'comment_prefix': '//',
        }
        run_kwargs.update(kwargs)

        stdout_capture = io.StringIO()
        with contextlib.redirect_stdout(stdout_capture):
            concat_all.concat_files(**run_kwargs)
        
        stdout_output = stdout_capture.getvalue()
        
        output_content = None
        if os.path.exists(output_file_path):
            with open(output_file_path, 'r', encoding='utf-8') as f:
                output_content = f.read()
                
        return output_content, stdout_output

class TestGitignoreHandling(BaseTestCase):
    def test_general_ignore(self):
        self._create_files({
            ".gitignore": "*.log\ntemp/",
            "file1.txt": "content1",
            "file2.log": "log_content",
            "temp/file3.txt": "temp_content",
            "another.txt": "content_another"
        })
        content, _ = self._run_concat_files(use_gitignore=True)
        self.assertIn("file1.txt", content)
        self.assertIn("content1", content)
        self.assertIn("another.txt", content)
        self.assertIn("content_another", content)
        self.assertNotIn("file2.log", content)
        self.assertNotIn("log_content", content)
        self.assertNotIn("temp/file3.txt", content)
        self.assertNotIn("temp_content", content)

    def test_negated_file_pattern(self):
        self._create_files({
            ".gitignore": "*.log\n!important.log",
            "file1.txt": "content1",
            "file2.log": "log_content_ignored",
            "important.log": "log_content_important" # Should be included despite *.log
        })
        # Test with log extension to try and pick up important.log
        content, _ = self._run_concat_files(extensions="txt,log", use_gitignore=True)
        self.assertIn("file1.txt", content)
        self.assertNotIn("file2.log", content)
        self.assertIn("important.log", content)
        self.assertIn("log_content_important", content)

    def test_negated_directory_pattern(self):
        self._create_files({
            ".gitignore": "temp/\n!temp/keep_this/",
            "file1.txt": "content1",
            "temp/ignored_dir/ignore.txt": "ignored_in_temp",
            "temp/keep_this/valuable.txt": "valuable_content",
            "temp/another_ignored.txt": "also_ignored"
        })
        content, _ = self._run_concat_files(use_gitignore=True)
        self.assertIn("file1.txt", content)
        self.assertNotIn("ignored_dir/ignore.txt", content)
        self.assertIn("temp/keep_this/valuable.txt", content) # Key assertion
        self.assertIn("valuable_content", content)
        self.assertNotIn("temp/another_ignored.txt", content)

    def test_negation_overrides_general_within_path(self):
        # Scenario: ignore all in 'build/', but not 'build/public/'
        self._create_files({
            ".gitignore": "build/\n!build/public/",
            "main.txt": "main content",
            "build/secret.txt": "secret",
            "build/public/index.txt": "public index"
        })
        content, _ = self._run_concat_files(use_gitignore=True)
        self.assertIn("main.txt", content)
        self.assertNotIn("build/secret.txt", content)
        self.assertIn("build/public/index.txt", content)
        self.assertIn("public index", content)


class TestDryRunMode(BaseTestCase):
    def test_dry_run_output_and_no_file_creation(self):
        self._create_files({
            "file1.py": "print('hello')",
            "subdir/file2.py": "print('world')"
        })
        output_content, stdout_output = self._run_concat_files(extensions="py", dry_run=True, output_file_name="dry_run_test.py")
        
        self.assertIsNone(output_content, "Output file should not be created in dry-run mode")
        
        self.assertIn("Dry run mode enabled.", stdout_output)
        expected_output_path = os.path.join(self.test_dir_path, "dry_run_test.py")
        self.assertIn(f"Output file would be: {expected_output_path}", stdout_output)
        
        # Normalize paths for comparison as os.walk order might vary slightly by OS
        # and our concat might pick them up in different order.
        # We care that both are mentioned.
        path1 = os.path.join(self.test_dir_path, "file1.py")
        path2 = os.path.join(self.test_dir_path, "subdir", "file2.py")
        self.assertIn(f"Would concatenate: {path1}", stdout_output)
        self.assertIn(f"Would concatenate: {path2}", stdout_output)

class TestExcludeOption(BaseTestCase):
    def test_exclude_specific_file(self):
        self._create_files({
            "file1.txt": "content1",
            "file_to_exclude.txt": "exclude_me",
            "file3.txt": "content3"
        })
        exclude_patterns = ["file_to_exclude.txt"]
        content, _ = self._run_concat_files(exclude_patterns=exclude_patterns)
        self.assertIn("file1.txt", content)
        self.assertNotIn("file_to_exclude.txt", content)
        self.assertIn("file3.txt", content)

    def test_exclude_by_glob_pattern(self):
        self._create_files({
            "file1.txt": "content1",
            "file.log": "log_data",
            "another.txt": "content_A",
            "another.log": "log_data_2"
        })
        exclude_patterns = ["*.log"]
        # Concatenate all files initially to check exclusion
        content, _ = self._run_concat_files(extensions="*", exclude_patterns=exclude_patterns)
        self.assertIn("file1.txt", content)
        self.assertNotIn("file.log", content)
        self.assertIn("another.txt", content)
        self.assertNotIn("another.log", content)

    def test_exclude_directory(self):
        self._create_files({
            "file1.txt": "content1",
            "build/file_in_build.txt": "build_content",
            "src/file_in_src.txt": "src_content",
            "build/another.txt": "build_content2"
        })
        exclude_patterns = ["build/"] # or "build/*" or "build"
        content, _ = self._run_concat_files(extensions="txt", exclude_patterns=exclude_patterns)
        self.assertIn("file1.txt", content)
        self.assertNotIn("build/file_in_build.txt", content)
        self.assertNotIn("build_content", content)
        self.assertIn("src/file_in_src.txt", content)

    def test_exclude_pattern_relative_to_root(self):
        self._create_files({
            "src/app1/main.py": "app1",
            "src/app2/main.py": "app2",
            "src/app1/utils.py": "app1_utils"
        })
        # Exclude all main.py files inside any subdirectory of src
        exclude_patterns = ["src/*/main.py"]
        content, _ = self._run_concat_files(extensions="py", exclude_patterns=exclude_patterns)
        self.assertNotIn("src/app1/main.py", content)
        self.assertNotIn("src/app2/main.py", content)
        self.assertIn("src/app1/utils.py", content)

class TestMaxDepthOption(BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_files({
            "root.txt": "root_content",
            "level1_dir/level1.txt": "level1_content",
            "level1_dir/level2_dir/level2.txt": "level2_content",
            "level1_dir/level2_dir/level3_dir/level3.txt": "level3_content",
            "another_level1_dir/another_level1.txt": "another_l1_content"
        })

    def test_max_depth_0(self):
        content, _ = self._run_concat_files(max_depth=0)
        self.assertIn("root.txt", content)
        self.assertNotIn("level1.txt", content)
        self.assertNotIn("level2.txt", content)
        self.assertNotIn("level3.txt", content)
        self.assertNotIn("another_level1.txt", content)

    def test_max_depth_1(self):
        content, _ = self._run_concat_files(max_depth=1)
        self.assertIn("root.txt", content)
        self.assertIn("level1.txt", content)
        self.assertIn("another_level1.txt", content)
        self.assertNotIn("level2.txt", content)
        self.assertNotIn("level3.txt", content)

    def test_max_depth_2(self):
        content, _ = self._run_concat_files(max_depth=2)
        self.assertIn("root.txt", content)
        self.assertIn("level1.txt", content)
        self.assertIn("another_level1.txt", content)
        self.assertIn("level2.txt", content)
        self.assertNotIn("level3.txt", content)
        
    def test_max_depth_unlimited(self): # default is -1
        content, _ = self._run_concat_files() # max_depth=-1 by default in concat_files
        self.assertIn("root.txt", content)
        self.assertIn("level1.txt", content)
        self.assertIn("another_level1.txt", content)
        self.assertIn("level2.txt", content)
        self.assertIn("level3.txt", content)

    def test_max_depth_stdout_message_dry_run(self):
        _, stdout = self._run_concat_files(max_depth=0, dry_run=True)
        # Path is relative to test_dir_path, which is "." for root in this context
        self.assertIn("Max depth (0) reached. Not traversing further from: .", stdout)

        _, stdout_l1 = self._run_concat_files(max_depth=1, dry_run=True)
        # Check for one of the level1 dirs
        path_l1d = os.path.join("level1_dir") 
        # The message should appear for each directory at max_depth
        self.assertIn(f"Max depth (1) reached. Not traversing further from: {path_l1d}", stdout_l1)
        path_al1d = os.path.join("another_level1_dir")
        self.assertIn(f"Max depth (1) reached. Not traversing further from: {path_al1d}", stdout_l1)


class TestMainOutputDir(BaseTestCase):
    def test_output_file_dir_with_default_name(self):
        self._create_files({"file1.txt": "content1"})
        output_dir = os.path.join(self.test_dir_path, "out")

        stdout_capture = io.StringIO()
        prev_argv = sys.argv
        sys.argv = ["concat-all", "txt", "-d", self.test_dir_path, "-D", output_dir]
        try:
            with contextlib.redirect_stdout(stdout_capture):
                concat_all.main()
        finally:
            sys.argv = prev_argv

        output_file_path = os.path.join(output_dir, "dump_txt.txt")
        self.assertTrue(os.path.exists(output_file_path))
        with open(output_file_path, "r", encoding="utf-8") as f:
            content = f.read()

        self.assertIn("file1.txt", content)


if __name__ == '__main__':
    unittest.main()



# Generic file concatenation utility

import os
import argparse
import fnmatch
import pathlib
import re
from datetime import datetime

def should_ignore_path(path, gitignore_patterns=None, base_dir=None):
    """Return True if ``path`` should be ignored according to ``gitignore_patterns``."""

    if not gitignore_patterns:
        return False

    path_str = str(pathlib.Path(path))
    abs_path = os.path.join(base_dir, path_str) if base_dir else path_str

    ignore = False
    for raw_pat in gitignore_patterns:
        negated = raw_pat.startswith('!')
        pat = raw_pat[1:] if negated else raw_pat
        if pat.startswith('/'):
            pat = pat[1:]

        if pat.endswith('/'):
            base_pat = pat[:-1]
            if (os.path.isdir(abs_path) and fnmatch.fnmatch(path_str, base_pat)) or \
               fnmatch.fnmatch(path_str, base_pat) or fnmatch.fnmatch(path_str, base_pat + '/*'):
                ignore = not negated
        else:
            if fnmatch.fnmatch(path_str, pat):
                ignore = not negated

    if ignore and os.path.isdir(abs_path):
        # If any negated pattern targets a path inside this directory, keep it
        for raw_pat in gitignore_patterns:
            if not raw_pat.startswith('!'):
                continue
            pat = raw_pat[1:]
            if pat.startswith('/'):
                pat = pat[1:]
            if pat.startswith(path_str.rstrip('/') + '/'):
                ignore = False
                break

    return ignore

def read_gitignore(dir_path):
    """
    Read .gitignore file and return list of patterns.
    
    Args:
        dir_path (str): Directory path where .gitignore is located
        
    Returns:
        list: List of gitignore patterns
    """
    gitignore_path = os.path.join(dir_path, '.gitignore')
    patterns = []
    
    if os.path.isfile(gitignore_path):
        with open(gitignore_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#'):
                    patterns.append(line)
                    
    return patterns

def concat_files(dir_path: str, file_extensions: str, output_file: str = './dump_{file_extension}.txt', 
                comment_prefix: str = '//', use_gitignore: bool = False, filename_suffix: str = '',
                dry_run: bool = False, exclude_patterns: list = None, exclude_regex_patterns: list = None,
                max_depth: int = -1):
    """
    Concatenate all files with the specified extension(s) under dir_path recursively
    and output to the specified output file.
    
    Args:
        dir_path (str): Directory path to search for files
        file_extensions (str or list): File extension(s) to filter (e.g., '.dart', '.tex')
                                      Can be a single string or a list of extensions
                                      Use '*' for all files
        output_file (str): Path to the output file
        comment_prefix (str): Prefix for comments in the output file
        use_gitignore (bool): Whether to filter files using .gitignore
        filename_suffix (str): Suffix to append to the output file name
        exclude_patterns (list): Glob patterns to exclude
        exclude_regex_patterns (list): Regex patterns to exclude
    """
    # Handle comma-separated extensions and remove leading dots and spaces
    extensions = [ext.strip('. ') for ext in file_extensions.split(',')]
    use_wildcard = '*' in extensions
    extensions_set = set(extensions)
    
    # Set output filename
    if use_wildcard:
        output_file = output_file.format(file_extension='all')
    else:
        output_file = output_file.format(file_extension='_'.join(extensions))

    if filename_suffix:
        if '{timestamp}' in filename_suffix:
            filename_suffix = filename_suffix.replace('{timestamp}', datetime.now().strftime('%Y%m%d_%H%M%S'))
        if '{unixtime}' in filename_suffix:
            filename_suffix = filename_suffix.replace('{unixtime}', str(int(datetime.now().timestamp())))
        base, ext = os.path.splitext(output_file)
        output_file = f"{base}{filename_suffix}{ext}"

    output_file_abs = os.path.abspath(output_file)
    if dry_run:
        print("Dry run mode enabled.")
        print(f"Output file would be: {output_file_abs}")
        output_handle = None
    else:
        if os.path.isfile(output_file_abs):
            os.remove(output_file_abs)
        output_handle = open(output_file_abs, 'a', encoding='utf-8')
    
    # Read gitignore if needed
    gitignore_patterns = []
    if use_gitignore:
        gitignore_patterns = read_gitignore(dir_path)
        # Add .git directory to the patterns when using gitignore
        if '.git' not in gitignore_patterns:
            gitignore_patterns.append('.git/')
        if '.gitignore' not in gitignore_patterns:
            gitignore_patterns.append('.gitignore')

    if exclude_patterns is None:
        exclude_patterns = []

    if exclude_regex_patterns is None:
        exclude_regex_patterns = []

    compiled_exclude_regex = [re.compile(pat) for pat in exclude_regex_patterns]
    
    dir_path = os.path.abspath(dir_path)
    # Calculate initial depth based on the absolute path
    # Normalize dir_path to remove any trailing slash for consistent depth calculation
    initial_depth = os.path.normpath(dir_path).count(os.sep)

    for root, dirs, files in os.walk(dir_path):
        current_abs_path = os.path.abspath(root)
        # Calculate current depth
        current_path_depth = os.path.normpath(current_abs_path).count(os.sep)
        relative_depth = current_path_depth - initial_depth

        # Apply max_depth logic
        if max_depth != -1 and relative_depth >= max_depth:
            if dry_run:
                print(f"Max depth ({max_depth}) reached. Not traversing further from: {os.path.relpath(root, dir_path) or '.'}")
            dirs[:] = [] # Prune subdirectories from further traversal

        current_rel_dir_path = os.path.relpath(root, dir_path)
        if current_rel_dir_path == ".": # Avoid "./" for root dir matching
            current_rel_dir_path = ""

        # Check if current directory should be ignored by gitignore (this should happen after depth check if depth is 0)
        # If max_depth is 0, we process files in root, then clear dirs. Gitignore on root itself still applies.
        if use_gitignore and should_ignore_path(current_rel_dir_path, gitignore_patterns, base_dir=dir_path):
            if dry_run:
                print(f"Ignoring directory (gitignore): {current_rel_dir_path if current_rel_dir_path else os.path.basename(root)}")
            dirs[:] = [] # Also prune if gitignored
            continue
        
        # Filter subdirectories based on --exclude patterns. This should also respect max_depth.
        # If dirs are already cleared by max_depth, this loop won't run or won't matter.
        original_dirs = list(dirs) 
        dirs[:] = [] 
        for d in original_dirs:
            dir_rel_path = os.path.join(current_rel_dir_path, d)
            excluded_by_custom_pattern = False

            for pattern in exclude_patterns:
                if fnmatch.fnmatch(dir_rel_path, pattern) or fnmatch.fnmatch(dir_rel_path + '/', pattern):
                    if dry_run:
                        print(f"Excluding directory (--exclude pattern '{pattern}'): {dir_rel_path}")
                    excluded_by_custom_pattern = True
                    break

            if not excluded_by_custom_pattern:
                for cregex in compiled_exclude_regex:
                    if cregex.search(dir_rel_path):
                        if dry_run:
                            print(f"Excluding directory (--exclude-regex '{cregex.pattern}'): {dir_rel_path}")
                        excluded_by_custom_pattern = True
                        break

            if not excluded_by_custom_pattern:
                dirs.append(d)
        
        for file in files:
            file_path = os.path.join(root, file)
            file_abs = file_path
            # Use current_rel_dir_path for file's relative path to ensure consistency
            rel_path = os.path.join(current_rel_dir_path, file)

            # Skip the output file itself
            if file_abs == output_file_abs:
                continue
                
            # Skip files matched by gitignore
            if use_gitignore and should_ignore_path(rel_path, gitignore_patterns, base_dir=dir_path):
                if dry_run: # Optional: log gitignore skips for files too
                    print(f"Ignoring file (gitignore): {rel_path}")
                continue

            # Skip files matched by --exclude patterns
            excluded_by_user_pattern = False
            for pattern in exclude_patterns:
                if fnmatch.fnmatch(rel_path, pattern):
                    if dry_run:
                        print(f"Excluding file (--exclude pattern '{pattern}'): {rel_path}")
                    else:
                        # Optional: print even if not dry_run, as per instructions
                        print(f"Excluding file (--exclude pattern '{pattern}'): {rel_path}")
                    excluded_by_user_pattern = True
                    break
            if excluded_by_user_pattern:
                continue

            for cregex in compiled_exclude_regex:
                if cregex.search(rel_path):
                    if dry_run:
                        print(f"Excluding file (--exclude-regex '{cregex.pattern}'): {rel_path}")
                    else:
                        print(f"Excluding file (--exclude-regex '{cregex.pattern}'): {rel_path}")
                    excluded_by_user_pattern = True
                    break
            if excluded_by_user_pattern:
                continue
                
            file_ext = os.path.splitext(file)[1]
            # Remove leading dot from file extension for comparison
            if file_ext.startswith('.'):
                file_ext = file_ext[1:]

            if use_wildcard or file_ext in extensions_set:
                if dry_run:
                    print(f"Would concatenate: {file_path}")
                else:
                    try:
                        with open(file_path, 'r', encoding='utf-8') as f:
                            content = f.read()
                        # Strip potential UTF-8 BOM to avoid leading artifacts in concatenated output
                        if content.startswith('\ufeff'):
                            content = content.lstrip('\ufeff')
                        output_handle.write(f"{comment_prefix} File: {file_path}\n{content}\n\n")
                    except (UnicodeDecodeError, PermissionError, IsADirectoryError):
                        # Skip binary files and files we can't read
                        pass

    if output_handle is not None:
        output_handle.close()

def main():
    # Example usage
    parser = argparse.ArgumentParser(description='Concatenate files with the specified extension(s) under the given directory.')
    parser.add_argument('file_extension', type=str, help='File extension(s) to filter (e.g., "txt", "dart" or "txt,dart,py"). Use "*" for all files. Leading dots are optional.')
    parser.add_argument('--dir_path', '-d', type=str, help='Path to the directory containing the files to concatenate')
    parser.add_argument('--output_file', '-o', type=str, help='Path to the output file')
    parser.add_argument('--output_file_dir', '-D', type=str, help='Path to the directory for the output file')
    parser.add_argument('--comment_prefix', '-c', type=str, help='Prefix for comments in the output file')
    parser.add_argument('--gitignore', '-i', action='store_true', help='Filter files using .gitignore')
    parser.add_argument('--filename_suffix', type=str, help='Suffix to append to the output filename (supports {timestamp}, {unixtime})')
    parser.add_argument('--dry-run', '-n', action='store_true', help='Enable dry run mode. Shows files that would be concatenated without actually writing them.')
    parser.add_argument('--exclude', type=str, help='Comma-separated list of glob patterns to exclude files/directories (e.g., "*.log,temp/*,specific_file.py")')
    parser.add_argument('--exclude-regex', type=str, help='Comma-separated list of regex patterns to exclude files/directories (matched against relative path)')
    parser.add_argument('--max-depth', type=int, default=-1, help='Maximum recursion depth. 0 for current directory, 1 for current + direct children, etc. Default: -1 (unlimited).')
    args = parser.parse_args()

    kwargs = {}
    kwargs['dir_path'] = args.dir_path if args.dir_path else os.getcwd()

    output_file = args.output_file
    if args.output_file_dir:
        os.makedirs(args.output_file_dir, exist_ok=True)
        if not output_file:
            output_file = 'dump_{file_extension}.txt'
        output_file = os.path.join(args.output_file_dir, output_file)

    if output_file:
        kwargs['output_file'] = output_file
    if args.comment_prefix:
        kwargs['comment_prefix'] = args.comment_prefix
    if args.gitignore:
        kwargs['use_gitignore'] = True
    if args.filename_suffix:
        kwargs['filename_suffix'] = args.filename_suffix
    if args.dry_run:
        kwargs['dry_run'] = True
    if args.exclude:
        kwargs['exclude_patterns'] = [p.strip() for p in args.exclude.split(',')]
    if args.exclude_regex:
        kwargs['exclude_regex_patterns'] = [p.strip() for p in args.exclude_regex.split(',') if p.strip()]
    if args.max_depth is not None: # Will always be true due to default, but good practice
        kwargs['max_depth'] = args.max_depth

    concat_files(dir_path=kwargs.pop('dir_path'), file_extensions=args.file_extension, **kwargs)
    
if __name__ == "__main__":
    main()

